//! Testing utilities and data for the augurs time series framework.
//!
//! Eventually I'd like this to be a fully fledged testing harness to automatically
//! compare results between the augurs, Python and R implementations, but for now
//! it's just a place to put some data.
#![warn(
    missing_docs,
    missing_debug_implementations,
    rust_2018_idioms,
    unreachable_pub
)]

pub mod data;
